@@ -1,11 +1,13 @@
-# -*- coding: utf-8 -*-
-
 """Test support of the various forms of tabular data."""
 
-from __future__ import print_function
-from __future__ import unicode_literals
 from tabulate import tabulate
-from common import assert_equal, assert_in, assert_raises, SkipTest
+from common import assert_equal, assert_in, raises, skip
+
+try:
+    from collections import UserDict
+except ImportError:
+    # Python2
+    from UserDict import UserDict
 
 
 def test_iterable_of_iterables():
@@ -123,8 +125,7 @@ def test_numpy_2d():
         result = tabulate(na, ["a", "b", "c"])
         assert_equal(expected, result)
     except ImportError:
-        print("test_numpy_2d is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_numpy_2d is skipped")
 
 
 def test_numpy_2d_firstrow():
@@ -139,8 +140,7 @@ def test_numpy_2d_firstrow():
         result = tabulate(na, headers="firstrow")
         assert_equal(expected, result)
     except ImportError:
-        print("test_numpy_2d_firstrow is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_numpy_2d_firstrow is skipped")
 
 
 def test_numpy_2d_keys():
@@ -161,8 +161,7 @@ def test_numpy_2d_keys():
         result = tabulate(na, headers="keys")
         assert_equal(expected, result)
     except ImportError:
-        print("test_numpy_2d_keys is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_numpy_2d_keys is skipped")
 
 
 def test_numpy_record_array():
@@ -188,8 +187,7 @@ def test_numpy_record_array():
         result = tabulate(na)
         assert_equal(expected, result)
     except ImportError:
-        print("test_numpy_2d_keys is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_numpy_2d_keys is skipped")
 
 
 def test_numpy_record_array_keys():
@@ -215,8 +213,7 @@ def test_numpy_record_array_keys():
         result = tabulate(na, headers="keys")
         assert_equal(expected, result)
     except ImportError:
-        print("test_numpy_2d_keys is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_numpy_2d_keys is skipped")
 
 
 def test_numpy_record_array_headers():
@@ -242,8 +239,7 @@ def test_numpy_record_array_headers():
         result = tabulate(na, headers=["person", "years", "cm"])
         assert_equal(expected, result)
     except ImportError:
-        print("test_numpy_2d_keys is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_numpy_2d_keys is skipped")
 
 
 def test_pandas():
@@ -263,8 +259,7 @@ def test_pandas():
         result = tabulate(df, headers=["string", "number"])
         assert_equal(expected, result)
     except ImportError:
-        print("test_pandas is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_pandas is skipped")
 
 
 def test_pandas_firstrow():
@@ -281,8 +276,7 @@ def test_pandas_firstrow():
         result = tabulate(df, headers="firstrow")
         assert_equal(expected, result)
     except ImportError:
-        print("test_pandas_firstrow is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_pandas_firstrow is skipped")
 
 
 def test_pandas_keys():
@@ -304,8 +298,7 @@ def test_pandas_keys():
         result = tabulate(df, headers="keys")
         assert_equal(expected, result)
     except ImportError:
-        print("test_pandas_keys is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_pandas_keys is skipped")
 
 
 def test_sqlite3():
@@ -327,8 +320,7 @@ Alice          23       169.5
 Bob            27       175"""
         assert_equal(expected, result)
     except ImportError:
-        print("test_sqlite3 is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_sqlite3 is skipped")
 
 
 def test_sqlite3_keys():
@@ -352,8 +344,7 @@ Alice          23       169.5
 Bob            27       175"""
         assert_equal(expected, result)
     except ImportError:
-        print("test_sqlite3_keys is skipped")
-        raise SkipTest()  # this test is optional
+        skip("test_sqlite3_keys is skipped")
 
 
 def test_list_of_namedtuples():
@@ -389,6 +380,15 @@ def test_list_of_dicts():
     assert_in(result, [expected1, expected2])
 
 
+def test_list_of_userdicts():
+    "Input: a list of UserDicts."
+    lod = [UserDict(foo=1, bar=2), UserDict(foo=3, bar=4)]
+    expected1 = "\n".join(["-  -", "1  2", "3  4", "-  -"])
+    expected2 = "\n".join(["-  -", "2  1", "4  3", "-  -"])
+    result = tabulate(lod)
+    assert_in(result, [expected1, expected2])
+
+
 def test_list_of_dicts_keys():
     "Input: a list of dictionaries, with keys as headers."
     lod = [{"foo": 1, "bar": 2}, {"foo": 3, "bar": 4}]
@@ -402,6 +402,19 @@ def test_list_of_dicts_keys():
     assert_in(result, [expected1, expected2])
 
 
+def test_list_of_userdicts_keys():
+    "Input: a list of UserDicts."
+    lod = [UserDict(foo=1, bar=2), UserDict(foo=3, bar=4)]
+    expected1 = "\n".join(
+        ["  foo    bar", "-----  -----", "    1      2", "    3      4"]
+    )
+    expected2 = "\n".join(
+        ["  bar    foo", "-----  -----", "    2      1", "    4      3"]
+    )
+    result = tabulate(lod, headers="keys")
+    assert_in(result, [expected1, expected2])
+
+
 def test_list_of_dicts_with_missing_keys():
     "Input: a list of dictionaries, with missing keys."
     lod = [{"foo": 1}, {"bar": 2}, {"foo": 4, "baz": 3}]
@@ -450,11 +463,11 @@ def test_list_of_dicts_with_list_of_headers():
     "Input: ValueError on a list of headers with a list of dicts (issue #23)"
     table = [{"letters": "ABCDE", "digits": 12345}]
     headers = ["DIGITS", "LETTERS"]
-    with assert_raises(ValueError):
+    with raises(ValueError):
         tabulate(table, headers=headers)
 
 
-def test_py27orlater_list_of_ordereddicts():
+def test_list_of_ordereddicts():
     "Input: a list of OrderedDicts."
     from collections import OrderedDict
 
@@ -463,3 +476,55 @@ def test_py27orlater_list_of_ordereddicts():
     expected = "\n".join(["  b    a", "---  ---", "  1    2", "  1    2"])
     result = tabulate(lod, headers="keys")
     assert_equal(expected, result)
+
+
+def test_py37orlater_list_of_dataclasses_keys():
+    "Input: a list of dataclasses with first item's fields as keys and headers"
+    try:
+        from dataclasses import make_dataclass
+
+        Person = make_dataclass("Person", ["name", "age", "height"])
+        ld = [Person("Alice", 23, 169.5), Person("Bob", 27, 175.0)]
+        result = tabulate(ld, headers="keys")
+        expected = "\n".join(
+            [
+                "name      age    height",
+                "------  -----  --------",
+                "Alice      23     169.5",
+                "Bob        27     175",
+            ]
+        )
+        assert_equal(expected, result)
+    except ImportError:
+        skip("test_py37orlater_list_of_dataclasses_keys is skipped")
+
+
+def test_py37orlater_list_of_dataclasses_headers():
+    "Input: a list of dataclasses with user-supplied headers"
+    try:
+        from dataclasses import make_dataclass
+
+        Person = make_dataclass("Person", ["name", "age", "height"])
+        ld = [Person("Alice", 23, 169.5), Person("Bob", 27, 175.0)]
+        result = tabulate(ld, headers=["person", "years", "cm"])
+        expected = "\n".join(
+            [
+                "person      years     cm",
+                "--------  -------  -----",
+                "Alice          23  169.5",
+                "Bob            27  175",
+            ]
+        )
+        assert_equal(expected, result)
+    except ImportError:
+        skip("test_py37orlater_list_of_dataclasses_headers is skipped")
+
+
+def test_list_bytes():
+    "Input: a list of bytes. (issue #192)"
+    lb = [["你好".encode("utf-8")], ["你好"]]
+    expected = "\n".join(
+        ["bytes", "---------------------------", r"b'\xe4\xbd\xa0\xe5\xa5\xbd'", "你好"]
+    )
+    result = tabulate(lb, headers=["bytes"])
+    assert_equal(expected, result)
